"""
A module implementing a sample task for UI testing purposes.
"""

# built-in
import asyncio
from random import random

# third-party
from runtimepy.net.arbiter import AppInfo
from runtimepy.net.arbiter.task import ArbiterTask, TaskFactory
from runtimepy.net.stream.json import JsonMessageConnection
from runtimepy.primitives import Uint8
from runtimepy.primitives.field import BitField, BitFlag


class SampleTask(ArbiterTask):
    """A base TUI application."""

    async def init(self, app: AppInfo) -> None:
        """Initialize this task with application information."""

        await super().init(app)

        # Register an enum.
        self.env.enum(
            "SampleEnum",
            "int",
            items={
                "zero": 0,
                "one": 1,
                "two": 2,
                "three": 3,
                "four": 4,
                "five": 5,
                "six": 6,
                "seven": 7,
                "eight": 8,
                "nine": 9,
                "ten": 10,
            },
        )

        self.env.enum(
            "InsanelyLongEnumNameForTesting",
            "int",
            {
                "very_long_member_name_0": 0,
                "very_long_member_name_1": 1,
                "very_long_member_name_2": 2,
                "very_long_member_name_3": 3,
            },
        )

        for name in ["a", "b", "c"]:
            with self.env.names_pushed(name):
                with self.env.names_pushed("fields"):
                    prim = Uint8()
                    self.env.int_channel("raw", prim)

                    # Add a bit field and flag.
                    self.env.add_field(BitFlag("flag1", prim, 0))
                    self.env.add_field(
                        BitFlag("flag2", prim, 1, commandable=True)
                    )
                    self.env.add_field(BitField("field1", prim, 2, 2))
                    self.env.add_field(BitField("field2", prim, 4, 4))

                for i in range(10):
                    with self.env.names_pushed(str(i)):
                        self.env.float_channel("random", "double")
                        self.env.int_channel("enum", enum="SampleEnum")
                        self.env.int_channel(
                            "really_really_long_enum",
                            enum="InsanelyLongEnumNameForTesting",
                        )
                        self.env.bool_channel("bool")
                        self.env.int_channel("int", commandable=True)
                        self.env.int_channel(
                            "scaled_int", commandable=True, scaling=[1.0, 2.0]
                        )
                        self.env.int_channel(
                            "scaled_float",
                            commandable=True,
                            scaling=[2.0, 3.0],
                        )

    async def dispatch(self) -> bool:
        """Dispatch an iteration of this task."""

        # Use this to implement / test rate-limited logging.
        with self.log_time("dispatch"):
            dispatches = self.metrics.dispatches.value % 10

            # Update local channels.
            for name in ["a", "b", "c"]:
                for i in range(10):
                    name_string = name + "." + str(i)
                    self.env.set(
                        f"{name_string}.enum", ((dispatches + i) % 10) + 1
                    )
                    self.env.set(f"{name_string}.bool", i % 2 == 0)

                    for chan in ["random"]:
                        self.env.set(f"{name_string}.{chan}", random())

            # Interact with connections.
            await asyncio.gather(
                *(
                    x.loopback()
                    for x in self.app.search(
                        pattern="client", kind=JsonMessageConnection
                    )
                )
            )

        return True


class Sample(TaskFactory[SampleTask]):
    """A TUI application factory."""

    kind = SampleTask



"""
An entry-point for the 'client' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from runtimepy.entry import main as runtimepy_main
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.io import ARBITER
from vcorelib.paths.context import tempfile

# internal
from conntextual import PKG_NAME
from conntextual.client import client_args, client_config
from conntextual.commands.common import (
    DEFAULT_VARIANT,
    common_cli_args,
    runtimepy_cli_args,
)


def client_cmd(args: _Namespace) -> int:
    """Execute the client command."""

    cli_args = runtimepy_cli_args(args)
    cli_args.append(f"package://{PKG_NAME}/{DEFAULT_VARIANT}.yaml")

    with tempfile(suffix=".json") as path:
        assert ARBITER.encode(path, client_config(args))[0]
        cli_args.append(str(path))

        print(f"runtimepy_main({cli_args})")
        result = runtimepy_main(cli_args)

    return result  # pragma: nocover


def add_client_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add client-command arguments to its parser."""

    common_cli_args(parser)
    client_args(parser)

    return client_cmd



"""
A module implementing a TUI application task.
"""

# built-in
import asyncio

# third-party
import psutil
from runtimepy.net.arbiter import AppInfo
from runtimepy.net.arbiter.task import ArbiterTask, TaskFactory

# internal
from conntextual.ui.base import Base


class TuiDispatchTask(ArbiterTask):
    """A class implementing a periodic task for a textual TUI."""

    tui: Base
    tui_task: asyncio.Task[None]
    process: psutil.Process

    def _add_housekeeping_metrics(self) -> None:
        """Initialize housekeeping metrics."""

        self.env.float_channel("memory_percent")
        self.env.float_channel("cpu_percent")

        self.env.bool_channel("update_table", commandable=True)
        self.env.bool_channel("update_log", commandable=True)
        self.env.bool_channel("update_plot", commandable=True)

        self.env.set("update_table", True)
        self.env.set("update_log", True)
        self.env.set("update_plot", True)

        self.process = psutil.Process()

    def poll_housekeeping(self) -> None:
        """Update housekeeping metrics."""

        self.env.set("memory_percent", psutil.virtual_memory().percent)

        with self.process.oneshot():
            self.env.set("cpu_percent", self.process.cpu_percent())

    async def init(self, app: AppInfo) -> None:
        """Initialize this task with application information."""

        await super().init(app)
        self.tui = Base.create(app, self.env)

        self._add_housekeeping_metrics()
        self.env.finalize()

        # Create application task.
        self.tui_task = asyncio.create_task(
            self.tui.run_async(
                headless=app.config.get("headless", False),  # type: ignore
            )
        )

        # Wait for the application to be composed.
        await self.tui.composed.wait()

    async def dispatch(self) -> bool:
        """Dispatch an iteration of this task."""

        self.poll_housekeeping()
        self.tui.dispatch(
            update_table=self.env.value("update_table"),  # type: ignore
            update_log=self.env.value("update_log"),  # type: ignore
            update_plot=self.env.value("update_plot"),  # type: ignore
        )

        return True

    async def stop_extra(self) -> None:
        """Extra actions to perform when this task is stopping."""

        # Ensure that the app task is awaited.
        await self.tui.action_quit()
        await self.tui_task


class TuiDispatch(TaskFactory[TuiDispatchTask]):
    """A factory for the TUI dispatch task."""

    kind = TuiDispatchTask



"""
A module implementing conntextual client interfaces.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from typing import Any

# third-party
from runtimepy.commands.common import arbiter_args


def client_args(
    parser: _ArgumentParser, default_factory: str = "tcp_json"
) -> None:
    """Add command-line argument options for servers."""

    parser.add_argument("host", help="hostname to connect to")
    parser.add_argument("port", type=int, help="port to connect to")
    parser.add_argument(
        "-f",
        "--factory",
        default=default_factory,
        help="connection factory to use (default: %(default)s)",
    )

    with arbiter_args(parser, nargs="*"):
        pass


def client_config(args: _Namespace) -> dict[str, Any]:
    """Get a server configuration based on command-line arguments."""

    result = {
        "includes": ["package://runtimepy/factories.yaml"],
        "clients": [
            {
                "name": "client",
                "factory": args.factory,
                "kwargs": {"host": args.host, "port": args.port},
            }
        ],
    }

    return result



"""
A module implementing a base application.
"""

# built-in
import asyncio
import curses

# third-party
from runtimepy.net.arbiter import AppInfo
from runtimepy.net.arbiter.task import ArbiterTask
from runtimepy.tui.cursor import Cursor
from runtimepy.tui.mixin import CursesWindow


class AppBase(ArbiterTask):
    """A base TUI application."""

    app: AppInfo
    input_queue: asyncio.Queue[int]

    async def init(self, app: AppInfo) -> None:
        """Initialize this task with application information."""

        self.app = app
        self.input_queue = asyncio.Queue()

        cursor = self.cursor

        with self.env.names_pushed("cursor"):
            self.env.channel("x", cursor.x)
            self.env.channel("y", cursor.y)

        with self.env.names_pushed("window"):
            self.env.channel("width", cursor.max_x)
            self.env.channel("height", cursor.max_y)

        self._handle_resize()

    def _handle_resize(self) -> None:
        """Handle the application getting re-sized."""

        self.cursor.poll_max()

    async def handle_char(self, char: int) -> None:
        """Handle user input."""

        if char == curses.KEY_RESIZE:
            self._handle_resize()
        else:  # pragma: nocover
            # trigger this with 'q'
            self.app.stop.set()

        # Handle this at some point.
        # elif char == curses.KEY_MOUSE:
        #     pass

    def draw(self) -> None:
        """Draw the application."""

    async def dispatch(self) -> bool:
        """Dispatch an iteration of this task."""

        window = self.window

        # Check for user input.
        keep_reading = True
        while keep_reading:
            data = window.getch()
            keep_reading = data != -1
            if keep_reading:
                self.input_queue.put_nowait(data)

        # Process inputs.
        while not self.input_queue.empty():
            await self.handle_char(self.input_queue.get_nowait())

        # Update state.
        self.draw()
        window.noutrefresh()
        curses.doupdate()

        return True

    @property
    def cursor(self) -> Cursor:
        """Get this instance's cursor."""
        return self.app.tui.cursor

    @property
    def window(self) -> CursesWindow:
        """Get this instance's window."""
        return self.app.tui.window



"""
A module implementing a curses-based user interface.
"""

# third-party
from runtimepy.net.arbiter import AppInfo
from runtimepy.net.arbiter.task import TaskFactory

# internal
from conntextual.curses.base import AppBase
from conntextual.curses.tui import Tui


class TuiApp(TaskFactory[Tui]):
    """A TUI application factory."""

    kind = Tui


async def run(app: AppInfo) -> int:
    """Run a textual application."""

    apps = list(app.search_tasks(AppBase))
    del apps

    return 0



# =====================================
# generator=datazen
# version=3.1.3
# hash=c09b8fc185aee1d63053118e6abd919e
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A network-application TUI using textual."
PKG_NAME = "conntextual"
VERSION = "1.3.0"



# =====================================
# generator=datazen
# version=3.1.3
# hash=f353fb494df76a27eff3f95cce1dd6b6
# =====================================

"""
A module aggregating package commands.
"""

# built-in
from typing import List as _List
from typing import Tuple as _Tuple

# third-party
from vcorelib.args import CommandRegister as _CommandRegister

# internal
from conntextual.commands.client import add_client_cmd
from conntextual.commands.ui import add_ui_cmd


def commands() -> _List[_Tuple[str, str, _CommandRegister]]:
    """Get this package's commands."""

    return [
        (
            "client",
            "attempt to connect a client to a remote session",
            add_client_cmd,
        ),
        (
            "ui",
            "run a user interface for runtimepy applications",
            add_ui_cmd,
        ),
        ("noop", "command stub (does nothing)", lambda _: lambda _: 0),
    ]



"""
A module implementing a data model for channel environments.
"""

# built-in
from dataclasses import dataclass
from enum import StrEnum

# third-party
from runtimepy.channel.environment import ChannelEnvironment
from runtimepy.channel.environment.command import ChannelCommandProcessor
from runtimepy.net.arbiter import AppInfo
from vcorelib.logging import LoggerType


class ChannelEnvironmentSource(StrEnum):
    """Possible sources of channel environments."""

    TASK = "task"
    CONNECTION_LOCAL = "local connection"
    CONNECTION_REMOTE = "remote connection"


@dataclass
class Model:
    """A model for channel environment displays."""

    name: str
    command: ChannelCommandProcessor
    source: ChannelEnvironmentSource
    logger: LoggerType
    app: AppInfo

    @property
    def env(self) -> ChannelEnvironment:
        """Get the channel environment."""
        return self.command.env



"""
A module impementing a channel-environment log widget.
"""

# built-in
from logging import ERROR, INFO, Formatter, Logger
from typing import Optional

# third-party
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Input, Log, Static
from vcorelib.logging import (
    DEFAULT_TIME_FORMAT,
    LoggerType,
    LogRecordQueue,
    queue_handler,
)

# internal
from conntextual.ui.channel.suggester import CommandSuggester

MAX_LINES = 1000


class InputWithHistory(Input):
    """An input with last-command history."""

    BINDINGS = [Binding("up", "previous_command", "previous command")]

    previous: str

    def action_previous_command(self) -> None:
        """Go back to the previous command."""

        if self.previous:
            self.value = self.previous
            self.action_end()
            self.refresh()


class ChannelEnvironmentLog(Static):
    """A channel-environment log widget."""

    parent_name: str
    logger: LoggerType
    queue: LogRecordQueue
    suggester: Optional[CommandSuggester]

    def dispatch(self) -> None:
        """Dispatch the log updater."""

        log = self.query_one(Log)
        while not self.queue.empty():
            log.write_line(self.queue.get_nowait().getMessage())

    @on(Input.Submitted)
    def handle_submit(self, event: Input.Submitted) -> None:
        """Handle input submission."""

        self.query_one(InputWithHistory).previous = event.value
        assert self.suggester is not None
        result = self.suggester.processor.command(event.value)

        self.logger.log(
            INFO if result else ERROR, "%s: %s", event.value, result
        )

        # Reset input.
        node = self.query_one(Input)
        node.action_home()
        node.action_delete_right_all()

    def compose(self) -> ComposeResult:
        """Create child nodes."""

        # Initialize logger handling.
        self.queue, handler = queue_handler(self.logger, root_formatter=False)
        handler.setFormatter(Formatter(DEFAULT_TIME_FORMAT))
        if self.logger is not Logger.root:
            self.logger.info("Queue handler initialized.")

        if self.suggester is not None:
            input_box = InputWithHistory(
                "set ",
                classes="command_input",
                suggester=self.suggester,
                id=f"{self.parent_name}-input",
            )
            input_box.previous = ""
            yield input_box

        yield Log(classes="log", max_lines=MAX_LINES)



# =====================================
# generator=datazen
# version=3.1.3
# hash=7c7cd6fb4915a6dbda6dfbc90aaf2fb5
# =====================================
"""
conntextual - Package's default entry-point.
"""

# built-in
import sys

# internal
from conntextual.entry import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
A module implementing color interfaces for channel enviroments.
"""

# built-in
from typing import Optional

# third-party
from runtimepy.enum import RuntimeEnum
from runtimepy.primitives.type import AnyPrimitiveType
from textual._color_constants import COLOR_NAME_TO_RGB
from textual.color import Color


def bit_field_style() -> str:
    """Get a style string for bit-field elements."""
    return Color(*COLOR_NAME_TO_RGB["teal"]).hex


def type_str_style(kind: AnyPrimitiveType, enum: Optional[RuntimeEnum]) -> str:
    """Get a style for a given type."""

    result = ""

    if kind.is_boolean:
        result = Color(*COLOR_NAME_TO_RGB["ansi_bright_cyan"]).hex
    elif kind.is_float:
        result = Color(*COLOR_NAME_TO_RGB["indigo"]).hex
    else:
        result = Color(*COLOR_NAME_TO_RGB["purple"]).hex

    if enum is not None:
        result += " bold"

    return result



"""
A module implementing an include/exclude pattern matching interface.
"""

# built-in
import re
from typing import List, NamedTuple, Optional, Union

PatternList = List[re.Pattern[str]]
StringOrList = Union[str, List[str]]


class PatternPair(NamedTuple):
    """A container for managing pattern data."""

    includes: PatternList
    excludes: PatternList

    def matches(self, data: str) -> bool:
        """Determine whether or not a string matches this pattern pair."""

        include_result = True

        if self.includes:
            include_result = any(
                include.search(data) is not None for include in self.includes
            )

        exclude_result = False

        if include_result:
            exclude_result = any(
                exclude.search(data) is not None for exclude in self.excludes
            )

        return include_result and not exclude_result

    @staticmethod
    def from_dict(data: dict[str, StringOrList]) -> "PatternPair":
        """Create a pattern pair from dictionary data."""

        includes: PatternList = []
        excludes: PatternList = []

        for patterns, keys in zip(
            [includes, excludes],
            [
                ["include", "includes", "included"],
                ["exclude", "excludes", "excluded"],
            ],
        ):
            for key in keys:
                pattern: Optional[StringOrList] = data.get(key)
                if pattern:
                    if isinstance(pattern, str):
                        patterns.append(re.compile(pattern))
                    else:
                        patterns += [re.compile(x) for x in pattern]

        return PatternPair(includes, excludes)



"""
Test that we can load test resources.
"""

# internal
from tests.resources import resource


def test_resources_basic():
    """Test that we can locate test data."""

    assert resource("test.txt").is_file()
    assert resource("test.txt", valid=False).is_file()



"""
An entry-point for the 'ui' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from runtimepy.commands.common import arbiter_args
from runtimepy.entry import main as runtimepy_main
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.io import ARBITER
from vcorelib.paths.context import tempfile

# internal
from conntextual import PKG_NAME
from conntextual.commands.common import (
    DEFAULT_VARIANT,
    common_cli_args,
    runtimepy_cli_args,
)
from conntextual.server import server_args, server_config


def ui_cmd(args: _Namespace) -> int:
    """Execute the ui command."""

    cli_args = runtimepy_cli_args(args)

    cli_args.append(f"package://{args.package}/{args.variant}.yaml")

    with tempfile(suffix=".json") as path:
        assert ARBITER.encode(path, server_config(args))[0]
        if not args.no_server:
            cli_args.append(str(path))

        print(f"runtimepy_main({cli_args})")
        result = runtimepy_main(cli_args)

    return result


def add_ui_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add ui-command arguments to its parser."""

    parser.add_argument(
        "-p",
        "--package",
        default=PKG_NAME,
        help="package to source application from",
    )
    parser.add_argument(
        "-v",
        "--variant",
        choices=[DEFAULT_VARIANT, "curses", "headless"],
        default=DEFAULT_VARIANT,
        help="application variant to use (default: %(default)s)",
    )
    common_cli_args(parser)
    server_args(parser)

    with arbiter_args(parser, nargs="*"):
        pass

    return ui_cmd



"""
A module implementing a user interface base application.
"""

# built-in
import asyncio
from contextlib import suppress
import logging
import os
from pathlib import Path
from typing import Optional

# third-party
from runtimepy.channel.environment import ChannelEnvironment
from runtimepy.net.arbiter import AppInfo
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.keys import Keys
from textual.logging import TextualHandler
from textual.widgets import Input, TabbedContent

# internal
from conntextual.ui.channel.environment import ChannelEnvironmentDisplay
from conntextual.ui.channel.model import ChannelEnvironmentSource
from conntextual.ui.channel.pattern import PatternPair
from conntextual.ui.footer import CustomFooter
from conntextual.ui.model import Model

TCSS_ROOT = Path(__file__).parent.parent.joinpath("data", "tcss")


class Base(App[None]):
    """A simple textual application."""

    BINDINGS = [
        ("q", "quit"),
        (Keys.Escape, "quit", "(or q) quit"),
        ("space", "toggle_pause", "toggle pause"),
        ("d", "toggle_dark", "toggle dark mode"),
        ("g", "screenshot", "take a screenshot"),
        ("r", "refresh_plot", "refresh plot"),
        ("R", "random_channel", "plot random channel"),
        Binding(Keys.Tab, "tab(True)", "Next tab", priority=True),
        Binding(Keys.BackTab, "tab(False)", "Previous tab", priority=True),
    ]

    CSS_PATH = TCSS_ROOT.joinpath("base.tcss")

    model: Model
    composed: asyncio.Event

    tab_pattern: PatternPair

    def action_toggle_pause(self) -> None:
        """Toggle pause state."""
        self.model.paused.toggle()

    def action_tab(self, forward: bool) -> None:
        """Change the active tab."""

        tabs = self.tabs
        curr = tabs.active
        if not curr:
            return

        # Accept an input box suggestion if an input box is highlighted.
        focus = self.focused
        if forward and focus is not None and isinstance(focus, Input):
            focus.action_cursor_right()
            return

        idx = int(curr.split("-")[1]) - 1

        idx = idx + 1 if forward else idx - 1

        num_tabs = len(self.model.environments)
        if idx >= num_tabs:
            idx -= num_tabs
        if idx < 0:
            idx += num_tabs

        tabs.active = f"tab-{idx + 1}"

        # Update footer.
        footer = self.query_one(CustomFooter)
        footer.current_tab = self.model.environments[idx].label
        footer.refresh()

    def compose_app(self) -> ComposeResult:
        """Application-specific interface creation."""

        footer = CustomFooter()
        footer.current_tab = self.model.environments[0].label
        yield footer

        with TabbedContent(
            *(x.model.name for x in self.model.environments), classes="tabs"
        ):
            yield from self.model.environments

    @property
    def tabs(self) -> TabbedContent:
        """Get the tab container."""
        return self.query_one(".tabs", expect_type=TabbedContent)

    def dispatch(
        self,
        update_table: bool = True,
        update_log: bool = True,
        update_plot: bool = True,
    ) -> None:
        """Update channel values."""

        self.model.uptime.value = (
            asyncio.get_running_loop().time() - self.model.start
        )

        if not self.model.paused:
            env = self.current_channel_environment
            if env is not None:
                env.update_channels(
                    update_table=update_table,
                    update_log=update_log,
                    update_plot=update_plot,
                )

    @property
    def current_channel_environment(
        self,
    ) -> Optional[ChannelEnvironmentDisplay]:
        """Get the current channel-environment display."""

        env = None

        with suppress(NoMatches):
            curr = self.tabs.active
            if curr:
                env = self.query_one(
                    f"#{self.model.tab_to_id[curr]}",
                    expect_type=ChannelEnvironmentDisplay,
                )

        return env

    def action_random_channel(self) -> None:
        """Randomize the channel on the current tab."""

        env = self.current_channel_environment
        if env is not None:
            env.random_channel()

    def action_refresh_plot(self) -> None:
        """Refresh the current plot."""

        env = self.current_channel_environment
        if env is not None:
            env.reset_plot()

    def ui_enabled(self, name: str) -> bool:
        """
        Determine if a task tab should be created for a task with a given name.
        """

        return name == "tui" or self.tab_pattern.matches(name)

    def _get_env_channel_pattern(self, name: str) -> PatternPair:
        """Get a channel-include pattern for a particular environment."""

        patterns = self.model.app.config.get("channel_patterns", {})
        return PatternPair.from_dict(patterns.get(name, {}))  # type: ignore

    def _init_environments(self) -> None:
        """Initialize channel-environment display instances."""

        # Channels for tasks and connections.
        self.model.environments += [
            ChannelEnvironmentDisplay.create(
                name,
                task.command,
                ChannelEnvironmentSource.TASK,
                task.logger,
                self.model.app,
                channel_pattern=self._get_env_channel_pattern(name),
            )
            for name, task in self.model.app.tasks.items()
            if self.ui_enabled(name)
        ] + [
            ChannelEnvironmentDisplay.create(
                name,
                conn.command,
                ChannelEnvironmentSource.CONNECTION_LOCAL,
                conn.logger,
                self.model.app,
                channel_pattern=self._get_env_channel_pattern(name),
            )
            for name, conn in self.model.app.connections.items()
            if self.ui_enabled(name)
        ]

        # Ensure the TUI task is always first.
        for idx, env in enumerate(self.model.environments):
            if env.model.name == "tui":
                self.model.environments[0], self.model.environments[idx] = (
                    env,
                    self.model.environments[0],
                )

        # One indexed tabs automatically enumerate for the tabbed environment,
        # keep a mapping of tabs index to element identifier.
        for idx, env in enumerate(self.model.environments):
            assert env.id is not None
            self.model.tab_to_id[f"tab-{1 + idx}"] = env.id

    def compose(self) -> ComposeResult:
        """Create child nodes."""

        self._init_environments()
        yield from self.compose_app()
        self.composed.set()

    @staticmethod
    def create(
        app: AppInfo, env: ChannelEnvironment, handle_debug: bool = True
    ) -> "Base":
        """Create an application instance."""

        if handle_debug and app.config.get("debug"):
            logging.basicConfig(level="NOTSET", handlers=[TextualHandler()])
            os.environ["TEXTUAL"] = "devtools,debug"

        result = Base()
        result.model = Model.create(app, env)
        result.composed = asyncio.Event()
        result.tab_pattern = PatternPair.from_dict(
            app.config.get("tab_pattern", {}),  # type: ignore
        )

        return result

    async def action_quit(self) -> None:
        """Stop the rest of the application when quitting."""
        await super().action_quit()
        self.model.app.stop.set()



"""
A module implementing a plot widget.
"""

# third-party
from numpy.typing import ArrayLike
from textual_plotext import PlotextPlot


class Plot(PlotextPlot):
    """A plot widget."""

    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        theme: str,
        marker: str,
        *args,
        title: str = "under construction",
        **kwargs,
    ) -> None:
        """Initialize this instance."""

        super().__init__(*args, **kwargs)
        self.x = x
        self.y = y
        self.title = title
        self.plot_theme = theme
        self.plot_marker = marker

    def on_show(self) -> None:
        """Handle showing the plot."""

        self.dispatch()

    def on_resize(self) -> None:
        """Handle re-size."""

        self.dispatch()

    def on_mount(self) -> None:
        """Initialize the plot."""

        self.update_title()
        self.plt.theme(self.plot_theme)

    def update_title(self, name: str = None) -> None:
        """Update the plot's title."""

        if name is not None:
            self.title = name

        self.plt.title(self.title)

    def dispatch(self) -> None:
        """Draw a new instance of the plot."""

        self.plt.clear_data()
        self.plt.plot(self.x, self.y, marker=self.plot_marker)  # type: ignore
        self.refresh()

    def set_data(self, x: ArrayLike, y: ArrayLike) -> None:
        """Assign new data."""

        self.x = x
        self.y = y
        self.dispatch()



"""
A module implementing a data model for base applications.
"""

# built-in
import asyncio
from dataclasses import dataclass
from typing import List

# third-party
from runtimepy.channel.environment import ChannelEnvironment
from runtimepy.mixins.environment import ChannelEnvironmentMixin
from runtimepy.net.arbiter import AppInfo
from runtimepy.primitives import Bool, Double

# internal
from conntextual.ui.channel.environment import ChannelEnvironmentDisplay


@dataclass
class Model(ChannelEnvironmentMixin):
    """A base application model."""

    app: AppInfo
    env: ChannelEnvironment
    environments: List[ChannelEnvironmentDisplay]

    uptime: Double
    paused: Bool
    start: float

    tab_to_id: dict[str, str]

    @staticmethod
    def create(app: AppInfo, env: ChannelEnvironment) -> "Model":
        """Create a model instance."""

        # Add environment channels.
        result = Model(
            app,
            env,
            [],
            Double(),
            Bool(),
            asyncio.get_running_loop().time(),
            {},
        )
        result.env.channel("uptime", result.uptime)

        return result



"""
A module implementing shared argument parsing interfaces.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from typing import List

# third-party
from vcorelib.logging import forward_flags

DEFAULT_VARIANT = "app"


def runtimepy_cli_args(args: _Namespace) -> List[str]:
    """Get base command-line arguments for a runtimepy invocation."""

    cli_args = ["runtimepy"]

    flags = set(forward_flags(args, ["curses", "verbose", "no_uvloop"]))

    is_headless = getattr(args, "variant", None) == "headless"

    if not args.verbose and not is_headless:
        flags.add("--quiet")

    cli_args.extend(flags)

    cli_args.append(args.cmd)
    cli_args.extend(list(forward_flags(args, ["init_only"])))

    # Ensure that the application continues to run when running the user
    # interface.
    if not is_headless or args.wait_for_stop:
        cli_args.append("--wait-for-stop")

    cli_args.extend(args.configs)

    return cli_args


def common_cli_args(parser: _ArgumentParser) -> None:
    """Add common command-line options."""

    parser.add_argument(
        "-c",
        "--cmd",
        default="arbiter",
        help="runtimepy command to run (default: %(default)s)",
    )



"""
A module implementing a basic user interface.
"""

# built-in
from asyncio import sleep
from dataclasses import dataclass

# third-party
from runtimepy.net.arbiter import AppInfo

# internal
from conntextual.ui.base import Base
from conntextual.ui.channel.environment import ChannelEnvironmentDisplay
from conntextual.ui.channel.log import ChannelEnvironmentLog, InputWithHistory
from conntextual.ui.channel.model import ChannelEnvironmentSource
from conntextual.ui.task import TuiDispatchTask

__all__ = [
    "Base",
    "test",
    "ChannelEnvironmentDisplay",
    "ChannelEnvironmentSource",
]


async def stop_after(app: AppInfo) -> int:
    """Run a textual application."""

    if "stop_after" in app.config:
        duration: float = app.config["stop_after"]  # type: ignore
        app.logger.info("Starting %f sleep.", duration)
        await sleep(duration)
        app.stop.set()
        app.logger.info("Set stop signal.")

    return 0


@dataclass
class MockEvent:
    """A mock event class."""

    value: str


@dataclass
class MockCoordinate:
    """A mock coordinate class."""

    row: int


@dataclass
class MockCellEvent:
    """A mock cell-selection event class."""

    coordinate: MockCoordinate


async def tui_test(tui: Base) -> None:
    """Test the UI."""

    await tui.composed.wait()

    iterations = 2 * len(tui.model.environments)

    # Cycle through tabs.
    for direction in [True, False]:
        for _ in range(iterations):
            tui.action_tab(direction)
            await sleep(0.05)

    tui.action_toggle_pause()
    tui.action_toggle_pause()

    # Test input tab handling.
    await tui.action_focus("tui-input")
    tui.action_tab(True)

    # Send some commands.
    for env in tui.model.environments:
        log = env.query_one(ChannelEnvironmentLog)
        input_box = env.query_one(InputWithHistory)

        assert log.suggester is not None
        processor = log.suggester.processor
        processor.parser.exit(message="null")

        processor.get_suggestion("set m")
        processor.get_suggestion("set e")

        env.handle_cell_selected(
            MockCellEvent(MockCoordinate(1)),  # type: ignore
        )

        tui.action_refresh_plot()
        tui.action_random_channel()

        for command in [
            "test",
            "help",
            "set a.0.random",
            "set a.0.random -f",
            "set a.0.random 0.5 -f",
            "set a.0.enum three -f",
            "set a.0.enum 2.8 -f",
            "set a.0.bool true -f",
            "set a.0.bool 1.1 -f",
            "toggle a.0.bool -f",
            "toggle a.0.bool -f",
            "toggle a.0.enum -f",
        ]:
            input_box.value = command
            log.handle_submit(MockEvent(command))  # type: ignore

            input_box.action_previous_command()

        await sleep(0.05)

    tui.model.app.stop.set()


async def test(app: AppInfo) -> int:
    """Run a textual application."""

    if not app.stop.is_set():
        periodics = list(app.search_tasks(kind=TuiDispatchTask))
        assert (
            len(periodics) == 1
        ), f"{len(periodics)} application tasks found!"

        await tui_test(periodics[0].tui)
        app.logger.info("Test complete, stopping application.")
        app.stop.set()

    return 0



"""
A module implementing an interface for channel selection.
"""

# built-in
from dataclasses import dataclass
from typing import List, Union

# third-party
from runtimepy.channel.environment.base import ChannelResult
from vcorelib.math.time import default_time_ns


@dataclass
class SelectedChannel:
    """A container for selected-channel information."""

    name: str
    channel: ChannelResult
    timestamps: List[float]
    values: List[Union[float, int, bool]]
    start_ns: int

    @staticmethod
    def create(name: str, channel: ChannelResult) -> "SelectedChannel":
        """Create a selected-channel instance."""

        return SelectedChannel(name, channel, [], [], default_time_ns())

    def reset(self) -> None:
        """Reset this channel's start time."""

        self.timestamps = []
        self.values = []
        self.start_ns = default_time_ns()

    def poll(self) -> None:
        """Poll the underlying channel."""

        chan = self.channel[0]
        last_updated = chan.raw.last_updated_ns

        # Need to handle not infinitely growing.
        if (not self.timestamps) or last_updated > self.timestamps[-1]:
            self.timestamps.append((last_updated - self.start_ns) / 1e9)
            self.values.append(chan.raw.scaled)



"""
A module implementing conntextual server interfaces.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
import socket
from typing import Any

# third-party
from runtimepy.net import IPv4Host, get_free_socket_name


def server_args(
    parser: _ArgumentParser,
    default_udp_factory: str = "udp_json",
    default_tcp_factory: str = "tcp_json",
    default_websocket_factory: str = "websocket_json",
    default_udp_port: int = 0,
    default_tcp_port: int = 0,
    default_websocket_port: int = 0,
) -> None:
    """Add command-line argument options for servers."""

    parser.add_argument(
        "--udp",
        default=default_udp_port,
        type=int,
        help="JSON message server UDP port (default: %(default)d)",
    )
    parser.add_argument(
        "--udp-factory",
        default=default_udp_factory,
        help="connection factory to use for UDP server (default: %(default)s)",
    )

    parser.add_argument(
        "--tcp",
        default=default_tcp_port,
        type=int,
        help="JSON message server TCP port (default: %(default)d)",
    )
    parser.add_argument(
        "--tcp-factory",
        default=default_tcp_factory,
        help="connection factory to use for UDP server (default: %(default)s)",
    )

    parser.add_argument(
        "--websocket",
        default=default_websocket_port,
        type=int,
        help="JSON message server WebSocket port (default: %(default)d)",
    )
    parser.add_argument(
        "--websocket-factory",
        default=default_websocket_factory,
        help=(
            "connection factory to use for WebSocket "
            "server (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "-n",
        "--no-server",
        action="store_true",
        help="don't start any JSON command listeners",
    )


def server_config(args: _Namespace) -> dict[str, Any]:
    """Get a server configuration based on command-line arguments."""

    config: dict[str, Any] = {
        "includes": ["package://runtimepy/factories.yaml"],
        "servers": [
            {"factory": args.tcp_factory, "kwargs": {"port": "$tcp_json"}},
            {
                "factory": args.websocket_factory,
                "kwargs": {"port": "$websocket_json", "host": "0.0.0.0"},
            },
        ],
        "port_overrides": {
            "tcp_json": get_free_socket_name(IPv4Host(port=args.tcp)).port,
            "websocket_json": get_free_socket_name(
                IPv4Host(port=args.websocket)
            ).port,
        },
    }

    if not args.init_only:
        config["clients"] = [
            {
                "factory": args.udp_factory,
                "name": "udp_json_server",
                "kwargs": {"local_addr": ["localhost", "$udp_json"]},
            }
        ]
        config["port_overrides"]["udp_json"] = get_free_socket_name(
            IPv4Host(port=args.udp), kind=socket.SOCK_DGRAM
        ).port

    return config



"""
A module implementing a custom footer widget.
"""

# built-in
from typing import Optional, cast

# third-party
from rich.console import RenderableType
from rich.text import Text
from textual.widgets import Footer


class CustomFooter(Footer):
    """An extension of the footer widget."""

    current_tab: Optional[str]

    def render(self) -> RenderableType:
        """Render the footer."""

        result: Text = cast(Text, super().render())

        # Manually add text for tab and shift tab.

        if self.current_tab:
            result = Text.assemble(
                result,
                "| ",
                Text(f"tab: {self.current_tab}", style="yellow bold"),
            )

        return result



# =====================================
# generator=datazen
# version=3.1.3
# hash=a13929149030d15ffef9467595f314b4
# =====================================

"""
conntextual - Package definition for distribution.
"""

# third-party
try:
    from setuptools_wrapper.setup import setup
except (ImportError, ModuleNotFoundError):
    from conntextual_bootstrap.setup import setup  # type: ignore

# internal
from conntextual import DESCRIPTION, PKG_NAME, VERSION

author_info = {
    "name": "Vaughn Kottler",
    "email": "vaughnkottler@gmail.com",
    "username": "vkottler",
}
pkg_info = {
    "name": PKG_NAME,
    "slug": PKG_NAME.replace("-", "_"),
    "version": VERSION,
    "description": DESCRIPTION,
    "versions": [
        "3.11",
        "3.12",
    ],
}
setup(
    pkg_info,
    author_info,
)



"""
A module implementing an input suggester for the channel-environment
command widget.
"""

# built-in
from typing import Optional

# third-party
from runtimepy.channel.environment.command import ChannelCommandProcessor
from textual.suggester import Suggester


class CommandSuggester(Suggester):
    """An input suggester for channel environment commands."""

    processor: ChannelCommandProcessor

    async def get_suggestion(self, value: str) -> Optional[str]:
        """Get an input suggestion."""
        return self.processor.get_suggestion(value)

    @staticmethod
    def create(processor: ChannelCommandProcessor) -> "CommandSuggester":
        """A method for creating a command suggester."""

        result = CommandSuggester()
        result.processor = processor
        return result



"""
A module implementing user interface elements for channel environments.
"""

# built-in
import random
from typing import Dict, List, Optional, Tuple, Union

# third-party
from rich.text import Text
from runtimepy.channel import AnyChannel
from runtimepy.channel.environment.command import ChannelCommandProcessor
from runtimepy.enum import RuntimeEnum
from runtimepy.net.arbiter import AppInfo
from runtimepy.registry.name import RegistryKey
from textual import on
from textual.app import ComposeResult
from textual.containers import HorizontalScroll, ScrollableContainer
from textual.coordinate import Coordinate
from textual.widgets import Collapsible, DataTable, Pretty, Static
from vcorelib.logging import LoggerType
from vcorelib.math import to_nanos

# internal
from conntextual.ui.channel.color import bit_field_style, type_str_style
from conntextual.ui.channel.log import ChannelEnvironmentLog
from conntextual.ui.channel.model import ChannelEnvironmentSource, Model
from conntextual.ui.channel.pattern import PatternPair
from conntextual.ui.channel.plot import Plot
from conntextual.ui.channel.selected import SelectedChannel
from conntextual.ui.channel.suggester import CommandSuggester

__all__ = ["ChannelEnvironmentDisplay"]
COLUMNS = ["type", "name", "value"]
DEFAULT_VALUE_COL_WIDTH = 22
STALE_THRESHOLD_NS = to_nanos(0.5)


def css_name(name: str) -> str:
    """Replace some characters that don't work in identifier values."""
    return name.replace(".", "_")


class ChannelEnvironmentDisplay(Static):
    """A channel-environment interface element."""

    model: Model

    by_index: List[Tuple[Coordinate, RegistryKey]]
    channels_by_row: Dict[int, SelectedChannel]

    selected: SelectedChannel
    row_idx: int

    channel_pattern: PatternPair

    def add_channel(
        self, name: str, chan: AnyChannel, enum: Optional[RuntimeEnum]
    ) -> int:
        """Add a channel to the table."""

        table = self.query_one(DataTable)
        env = self.model.env

        self.channels_by_row[self.row_idx] = SelectedChannel.create(
            name, (chan, enum)
        )

        kind_str = str(chan.type)

        # Should handle enums at some point.
        if enum is not None:
            enum_name = env.enums.names.name(enum.id)
            assert enum_name is not None
            kind_str = enum_name

        table.add_row(
            Text(kind_str, style=type_str_style(chan.type, enum)),
            name if not chan.commandable else Text(name, style="bold green"),
            " " * max(len(str(env.value(name))), DEFAULT_VALUE_COL_WIDTH),
        )
        return chan.id

    def add_field(self, name: str) -> None:
        """Add a bit-field row entry."""

        table = self.query_one(DataTable)
        env = self.model.env

        field = env.fields[name]

        table.add_row(
            Text(
                f"{'bit' if field.width == 1 else 'bits'} {field.where_str()}",
                style=bit_field_style(),
            ),
            name if not field.commandable else Text(name, style="bold green"),
            " " * max(len(str(env.value(name))), DEFAULT_VALUE_COL_WIDTH),
        )

    def on_mount(self) -> None:
        """Populate channel table."""

        table = self.query_one(DataTable)
        env = self.model.env
        assert env.finalized
        names = list(env.names)

        # Set up columns.
        table.add_columns(*COLUMNS)
        val_col = COLUMNS.index("value")

        ident: RegistryKey
        for name in names:
            if not self.channel_pattern.matches(name):
                continue

            # Add channel rows.
            chan_result = env.get(name)
            if chan_result is not None:
                chan, enum = chan_result
                ident = self.add_channel(name, chan, enum)

            # Add field and flag rows.
            else:
                self.add_field(name)
                ident = name

            self.by_index.append((Coordinate(self.row_idx, val_col), ident))
            self.row_idx += 1

    def switch_to_channel(self, row: int) -> None:
        """Switch the plot to a channel at the specified row."""

        if row in self.channels_by_row:
            # Select channel.
            self.selected = self.channels_by_row[row]

            # Update plot parameters.
            name = self.selected.name
            self.query_one(Plot).update_title(name=name)
            self.model.logger.info("Switched plot to channel '%s'.", name)
            self.reset_plot()

    def random_channel(self) -> None:
        """Switch to a random channel."""

        row = -1
        while row not in self.channels_by_row:
            row = random.randint(0, self.row_idx - 1)

        self.switch_to_channel(row)

    def reset_plot(self) -> None:
        """Reset the selected plot."""

        self.selected.reset()
        self.query_one(Plot).set_data(
            self.selected.timestamps, self.selected.values
        )
        self.model.logger.info("Plot reset.")

    @on(DataTable.CellSelected)
    def handle_cell_selected(self, event: DataTable.CellSelected) -> None:
        """Handle input submission."""

        self.switch_to_channel(event.coordinate.row)

    def update_channels(
        self,
        update_table: bool = True,
        update_log: bool = True,
        update_plot: bool = True,
    ) -> None:
        """Update all channel values."""

        env = self.model.env

        if update_table:
            table = self.query_one(DataTable)

            for coord, chan in self.by_index:
                val = env.value(chan)
                if isinstance(val, float):
                    val = f"{val: 15.6f}"
                elif isinstance(val, bool):
                    val = "true" if val else "false"
                elif isinstance(val, int):
                    val = f"{val: 8d}       "

                # Get the age of the primitive.
                age = env.age_ns(chan)
                if age > STALE_THRESHOLD_NS:
                    val = Text(val, style="yellow")  # type: ignore

                table.update_cell_at(coord, val)

        # Update logs.
        if update_log:
            self.query_one(ChannelEnvironmentLog).dispatch()

        # Update plot.
        if update_plot:
            self.selected.poll()
            self.query_one(Plot).dispatch()

    @property
    def label(self) -> str:
        """Obtain a label string for this instance."""
        return f"({self.model.source}) {self.model.name}"

    def compose(self) -> ComposeResult:
        """Create child nodes."""

        with HorizontalScroll(classes="channels"):
            yield DataTable[Union[str, int, float]]()

            yield Plot(
                self.selected.timestamps,
                self.selected.values,
                str(self.model.app.config.get("plot_theme", "pro")),
                str(self.model.app.config.get("plot_marker", "braille")),
                title=self.selected.name,
                id="plot",
            )

        # Create log and command widget.
        log = ChannelEnvironmentLog()
        log.parent_name = self.model.name
        log.logger = self.model.logger
        log.suggester = CommandSuggester.create(self.model.command)
        yield log

        with ScrollableContainer():
            with Collapsible(title="configuration"):
                yield Pretty(self.model.app.config.get("root", {}))

    @staticmethod
    def create(
        name: str,
        command: ChannelCommandProcessor,
        source: ChannelEnvironmentSource,
        logger: LoggerType,
        app: AppInfo,
        channel_pattern: PatternPair,
    ) -> "ChannelEnvironmentDisplay":
        """Create a channel-environment display."""

        result = ChannelEnvironmentDisplay(id=css_name(name))
        result.model = Model(name, command, source, logger, app)
        result.by_index = []
        result.channels_by_row = {}
        result.row_idx = 0
        result.channel_pattern = channel_pattern

        names = list(result.model.env.names)
        assert names

        chan = None
        while chan is None:
            name = random.choice(names)
            chan = result.model.env.get(name)

        result.selected = SelectedChannel.create(name, chan)

        return result



"""
A module implementing a basic TUI application.
"""

# third-party
from runtimepy.channel.environment.command import ENVIRONMENTS

# internal
from conntextual.curses.base import AppBase


class Tui(AppBase):
    """A simple TUI application."""

    def _handle_resize(self) -> None:
        """Handle the application getting re-sized."""

        super()._handle_resize()
        self.cursor.reset()

        # Draw tabs.
        # self.tabs = self.window.derwin(
        #     3, self.env.value("window.width"), 0, 0
        # )
        # self.tabs.border()
        # need a cursor for this thing

        # self.tabs.addstr(1, 1, "test")

    def draw(self) -> None:
        """Draw the application."""

        self.cursor.reset()

        window = self.window

        for env_name, env in ENVIRONMENTS.items():
            window.addstr(f"========== {env_name} ==========")
            window.clrtoeol()
            if not self.cursor.inc_y():
                break

            for name in env.env.names:
                line = name

                chan, enum = env.env[name]

                # do floats better
                line += " " + str(chan)

                # Handle enum at some point.
                del enum
                # if enum is not None:
                #     pass

                window.addstr(line)
                window.clrtoeol()
                if not self.cursor.inc_y():
                    break
